// This code example is created for educational purpose
// by Thorsten Thormaehlen (contact: www.thormae.de).
// It is distributed without any warranty.

package de.unimarburg.shadertextureapp;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import android.opengl.GLES20;
import android.opengl.GLSurfaceView;
import android.opengl.GLUtils;
import android.app.Activity;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Matrix;
import android.os.Bundle;
import android.os.SystemClock;
import android.util.Log;

class Renderer {

  public float t;
  
  private String TAG = "de.unimarburg.shadertextureapp";
  private Context appContext;
  private enum VBOs {PyramidAll, numVBOs};
  private int[] bufID = new int[VBOs.numVBOs.ordinal()];
  private int pyramidVertNo = 0;
  private int texID = 0;
  private int progID = 0;
  private int vertID = 0;
  private int fragID = 0;
  private int vertexLoc = 0;
  private int colorLoc = 0;
  private int texCoordLoc = 0;
  private int normalLoc = 0;
  private int projectionLoc = 0;
  private int modelviewLoc = 0;
  private int texLoc = 0;
  private float[] projection = new float[16];
  private float[] modelview = new float[16];
  
  private int SIZEOF_FLOAT = Float.SIZE / 8;
  
  public Renderer(Context context) {
    appContext = context;
  }
  
  public void init(GL10 unused, EGLConfig config) {
    GLES20.glEnable(GLES20.GL_DEPTH_TEST);

    setupShaders();

    // generate a Vertex Buffer Object (VBO)
    GLES20.glGenBuffers(VBOs.numVBOs.ordinal(), bufID, 0);

    float pyramidVertexData[] = {
        0.0f, 0.0f, 2.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.5f, 1.0f, 0.0000f,-0.9701f, 0.2425f,
       -0.5f,-0.5f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0000f,-0.9701f, 0.2425f,
        0.5f,-0.5f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.0000f,-0.9701f, 0.2425f,
        0.0f, 0.0f, 2.0f, 0.0f, 1.0f, 0.0f, 1.0f, 0.5f, 1.0f, 0.9701f, 0.0000f, 0.2425f,
        0.5f,-0.5f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.9701f, 0.0000f, 0.2425f,
        0.5f, 0.5f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.9701f, 0.0000f, 0.2425f,
        0.0f, 0.0f, 2.0f, 0.0f, 0.0f, 1.0f, 1.0f, 0.5f, 1.0f, 0.0000f, 0.9701f, 0.2425f,
        0.5f, 0.5f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.0f, 0.0000f, 0.9701f, 0.2425f,
       -0.5f, 0.5f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f, 0.0f, 0.0000f, 0.9701f, 0.2425f,
        0.0f, 0.0f, 2.0f, 1.0f, 1.0f, 0.0f, 1.0f, 0.5f, 1.0f,-0.9701f, 0.0000f, 0.2425f,
       -0.5f, 0.5f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f, 0.0f, 0.0f,-0.9701f, 0.0000f, 0.2425f,
       -0.5f,-0.5f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 0.0f,-0.9701f, 0.0000f, 0.2425f
    }; 
    
    pyramidVertNo = 12;

    int floatItems = pyramidVertNo * (3+4+2+3);
    ByteBuffer bb = ByteBuffer.allocateDirect(floatItems * SIZEOF_FLOAT);
    bb.order(ByteOrder.nativeOrder());
    FloatBuffer pyramidVertexFB = bb.asFloatBuffer();
    pyramidVertexFB.put(pyramidVertexData);
    pyramidVertexFB.flip();

    GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, bufID[VBOs.PyramidAll.ordinal()]);
    GLES20.glBufferData(GLES20.GL_ARRAY_BUFFER, pyramidVertexFB.capacity()*SIZEOF_FLOAT, 
        pyramidVertexFB, GLES20.GL_STATIC_DRAW);
    
    texID = loadTexture(R.drawable.checkerboard);
  }
  
  public void resize(GL10 unused, int width, int height) {
    GLES20.glViewport(0, 0, width, height);
    
    // this function replaces gluPerspective
    mat4Perspective(projection, 30.0f, (float)width/(float)height, 1.0f, 10.0f);
    //mat4Print(projection);
  }

  
  public void display(GL10 unused) {

    GLES20.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);

    // camera orbits in the z=5 plane
    // and looks at the origin
    // mat4LookAt replaces gluLookAt
    double rad = Math.PI / 180.0f * t;
    mat4LookAt(modelview,
        5.0f*(float)Math.cos(rad), 5.0f*(float)Math.sin(rad), 5.0f, // eye
        0.0f, 0.0f, 0.5f, // look at
        0.0f, 0.0f, 1.0f); // up
    
    GLES20.glUseProgram(progID); 
    
    // load the current projection and modelview matrix into the
    // corresponding UNIFORM variables of the shader
    GLES20.glUniformMatrix4fv(projectionLoc, 1, false, projection, 0);
    GLES20.glUniformMatrix4fv(modelviewLoc, 1, false, modelview, 0);
    
    // activate texture unit 0
    GLES20.glActiveTexture(GLES20.GL_TEXTURE0);
    // bind texture
    GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, texID);
    // inform the shader to use texture unit 0
    GLES20.glUniform1i(texLoc, 0);
    
    // bind Triangle VBO
    GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, bufID[VBOs.PyramidAll.ordinal()]);
    int stride = (3 + 4 + 2 + 3) * SIZEOF_FLOAT;
    int offset = 0;

    // position
    if (vertexLoc != -1) {
      GLES20.glVertexAttribPointer(vertexLoc, 3, GLES20.GL_FLOAT, false, stride,
          offset);
      GLES20.glEnableVertexAttribArray(vertexLoc);
    }

    // color
    if (colorLoc != -1) {
      offset = 0 + 3 * SIZEOF_FLOAT;
      GLES20.glVertexAttribPointer(colorLoc, 4, GLES20.GL_FLOAT, false, stride, offset);
      GLES20.glEnableVertexAttribArray(colorLoc);
    }

    // texCoord
    if (texCoordLoc != -1) {
      offset = 0 + (3 + 4) * SIZEOF_FLOAT;
      GLES20.glVertexAttribPointer(texCoordLoc, 2, GLES20.GL_FLOAT, false, stride,
          offset);
      GLES20.glEnableVertexAttribArray(texCoordLoc);
    }

    // normal
    if (normalLoc != -1) {
      offset = 0 + (3 + 4 + 2) * SIZEOF_FLOAT;
      GLES20.glVertexAttribPointer(normalLoc, 3, GLES20.GL_FLOAT, false, stride,
          offset);
      GLES20.glEnableVertexAttribArray(normalLoc);
    }
    
    // render data
    GLES20.glDrawArrays(GLES20.GL_TRIANGLES, 0, pyramidVertNo);

    GLES20.glFlush();
  }
  
 
  
  public void  setupShaders() {
     
    vertID = GLES20.glCreateShader(GLES20.GL_VERTEX_SHADER);
    fragID = GLES20.glCreateShader(GLES20.GL_FRAGMENT_SHADER);
     
    String vs = loadShaderSrc(R.raw.texture_vert); 
    String fs = loadShaderSrc(R.raw.texture_frag);
   
    GLES20.glShaderSource(vertID, vs);
    GLES20.glShaderSource(fragID, fs);

    // compile the shader
    GLES20.glCompileShader(vertID);
    GLES20.glCompileShader(fragID);

    // check for errors
    printShaderInfoLog(vertID);
    printShaderInfoLog(fragID);

    // create program and attach shaders
    progID = GLES20.glCreateProgram();
    GLES20.glAttachShader(progID, vertID);
    GLES20.glAttachShader(progID, fragID);

    // link the program
    GLES20.glLinkProgram(progID);
    // output error messages
    printProgramInfoLog(progID);

    // "inputPosition" and "inputColor" are user-provided
    // IN variables of the vertex shader.
    // Their locations are stored to be used later with
    // glEnableVertexAttribArray()
    vertexLoc = GLES20.glGetAttribLocation(progID,"inputPosition");
    colorLoc = GLES20.glGetAttribLocation(progID, "inputColor");
    texCoordLoc = GLES20.glGetAttribLocation(progID,"inputTexCoord");
    normalLoc = GLES20.glGetAttribLocation(progID, "inputNormal");
    
    // "projection" and "modelview" are user-provided
    // UNIFORM variables of the vertex shader.
    // Their locations are stored to be used later
    projectionLoc = GLES20.glGetUniformLocation(progID, "projection");
    modelviewLoc = GLES20.glGetUniformLocation(progID, "modelview");
    texLoc = GLES20.glGetUniformLocation(progID, "myTexture");
  
  }
  
  private String loadShaderSrc(int resourceId) {
    StringBuilder sb = new StringBuilder();
    try {
      InputStream is = appContext.getResources().openRawResource(resourceId);
      BufferedReader br = new BufferedReader(new InputStreamReader(is));
      String line;
      while ((line = br.readLine()) != null) {
        sb.append(line);
        sb.append('\n');
      }
      br.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return sb.toString();
  }
 
  private void printShaderInfoLog(int obj) {
    IntBuffer infoLogLengthBuf = IntBuffer.allocate(1);
    GLES20.glGetShaderiv(obj, GLES20.GL_INFO_LOG_LENGTH, infoLogLengthBuf);
    int infoLogLength = infoLogLengthBuf.get(0);
    if (infoLogLength > 0) {
      Log.e(TAG, GLES20.glGetShaderInfoLog(obj));
    }
  }
  
  private void printProgramInfoLog(int obj) {
    IntBuffer infoLogLengthBuf = IntBuffer.allocate(1);
    GLES20.glGetProgramiv(obj, GLES20.GL_INFO_LOG_LENGTH, infoLogLengthBuf);
    int infoLogLength = infoLogLengthBuf.get(0);
    if (infoLogLength > 0) {
      Log.e(TAG, GLES20.glGetProgramInfoLog(obj));
    }
  }

  // returns a valid textureID on success, otherwise 0
  private int loadTexture(int resourceId) {

    int level = 0;
    int border = 0;

    // read the image from the resources using BitmapFactory
    BitmapFactory.Options options = new BitmapFactory.Options();
    options.inScaled = false; // No scaling
    Bitmap bitmap = BitmapFactory.decodeResource(appContext.getResources(),
        resourceId, options);

    // we need to vertically flip the image because the image origin
    // in OpenGL is the lower-left corner.
    Matrix flipMat = new Matrix();
    flipMat.setScale(1.0f, -1.0f);
    Bitmap flippedBitmap = Bitmap.createBitmap(bitmap, 0, 0, bitmap.getWidth(),
        bitmap.getHeight(), flipMat, false);
    bitmap.recycle();

    // request textureID
    final int[] textureID = new int[1];
    GLES20.glGenTextures(1, textureID, 0);
    if (textureID[0] == 0)
      return 0;

    // bind texture
    GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textureID[0]);

    // define how to filter the texture
    GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MAG_FILTER,
        GLES20.GL_LINEAR);
    GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MIN_FILTER,
        GLES20.GL_LINEAR);

    // specify the 2D texture map
    GLUtils.texImage2D(GLES20.GL_TEXTURE_2D, level, flippedBitmap, border);

    // the texture data is now handled by OpenGL, we can free the local copy
    flippedBitmap.recycle();

    return textureID[0];

  }
  
  // the following functions are some matrix and vector helpers
  // they work for this demo but in general it is recommended
  // to use more advanced matrix libraries
  private float vec3Dot(float[] a, float[] b) {
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
  }

  private void vec3Cross(float[] a, float[] b, float[] res) {
    res[0] = a[1] * b[2]  -  b[1] * a[2];
    res[1] = a[2] * b[0]  -  b[2] * a[0];
    res[2] = a[0] * b[1]  -  b[0] * a[1];
  }

  private void vec3Normalize(float[] a) {
    float mag = (float) Math.sqrt(a[0] * a[0]  +  a[1] * a[1]  +  a[2] * a[2]);
    a[0] /= mag; a[1] /= mag; a[2] /= mag;
  } 
  
  private void mat4Identity(float[] a) {
    for (int i = 0; i < 16; ++i) a[i] = 0.0f;
    for (int i = 0; i < 4; ++i) a[i + i * 4] = 1.0f;
  }

  private void mat4Multiply(float[] a, float[] b, float[] res) {
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        res[j*4 + i] = 0.0f;
        for (int k = 0; k < 4; ++k) {
          res[j*4 + i] += a[k*4 + i] * b[j*4 + k];
        }
      }
    }
  }
  
  private void mat4Perspective(float[] a, float fov, float aspect, float zNear, float zFar) {
    float f = 1.0f / (float) (Math.tan (fov/2.0f * (Math.PI / 180.0f)));
    mat4Identity(a);
    a[0] = f / aspect;
    a[1 * 4 + 1] = f;
    a[2 * 4 + 2] = (zFar + zNear)  / (zNear - zFar);
    a[3 * 4 + 2] = (2.0f * zFar * zNear) / (zNear - zFar);
    a[2 * 4 + 3] = -1.0f;
    a[3 * 4 + 3] = 0.0f;
  }
  
  private void mat4LookAt(float[] viewMatrix,
      float eyeX, float eyeY, float eyeZ,
      float centerX, float centerY, float centerZ,
      float upX, float upY, float upZ) {

    float dir[] = new float[3];
    float right[] = new float[3];
    float up[] = new float[3];
    float eye[] = new float[3];

    up[0]=upX; up[1]=upY; up[2]=upZ;
    eye[0]=eyeX; eye[1]=eyeY; eye[2]=eyeZ;

    dir[0]=centerX-eyeX; dir[1]=centerY-eyeY; dir[2]=centerZ-eyeZ;
    vec3Normalize(dir);
    vec3Cross(dir,up,right);
    vec3Normalize(right);
    vec3Cross(right,dir,up);
    vec3Normalize(up);
    // first row
    viewMatrix[0]  = right[0];
    viewMatrix[4]  = right[1];
    viewMatrix[8]  = right[2];
    viewMatrix[12] = -vec3Dot(right, eye);
    // second row
    viewMatrix[1]  = up[0];
    viewMatrix[5]  = up[1];
    viewMatrix[9]  = up[2];
    viewMatrix[13] = -vec3Dot(up, eye);
    // third row
    viewMatrix[2]  = -dir[0];
    viewMatrix[6]  = -dir[1];
    viewMatrix[10] = -dir[2];
    viewMatrix[14] =  vec3Dot(dir, eye);
    // forth row
    viewMatrix[3]  = 0.0f;
    viewMatrix[7]  = 0.0f;
    viewMatrix[11] = 0.0f;
    viewMatrix[15] = 1.0f;
  }
  
  private void mat4Print(float[] a) {
    // opengl uses column major order
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        System.out.print( a[j * 4 + i] + " ");
      }
      System.out.println(" ");
    }
  }

}

class GLSurfaceViewRenderer implements GLSurfaceView.Renderer {

  private Renderer renderer;
  
  final int targetFPS = 60;
  final int targetMillis = 1000 / targetFPS;
  private long startTime = 0;
  
  public GLSurfaceViewRenderer(Context appContext) {
    renderer = new Renderer(appContext);
  }
  
  @Override
  public void onDrawFrame(GL10 gl10) {
    float offset = 1.0f;
    renderer.t += offset;
    renderer.display(gl10);
    
    // this code attempts to achieve a target frames-per-second rate
    long timeTillNextRender = startTime + targetMillis;
    long sleepTime = timeTillNextRender - SystemClock.uptimeMillis();
    if (sleepTime > 0) {
      try {
        Thread.sleep(sleepTime);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    } else {
      // rendering is too slow
    }
    startTime = SystemClock.uptimeMillis();
  }

  @Override
  public void onSurfaceChanged(GL10 gl10, int width, int height) {
    renderer.resize(gl10, width, height);
  }

  @Override
  public void onSurfaceCreated(GL10 gl10, EGLConfig config) {
    renderer.init(gl10, config);
    startTime = SystemClock.uptimeMillis();
  }

  public Renderer getRenderer() {
    return renderer;
  }
}

public class ShaderTexture extends Activity {

  private GLSurfaceView myGLSurfaceView;
  private GLSurfaceViewRenderer myGLSurfaceViewRenderer;

  @Override
  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);
    myGLSurfaceView = new GLSurfaceView(this);
    
    // Create an OpenGL ES 2.0 context.
    myGLSurfaceView.setEGLContextClientVersion(2);
    
    myGLSurfaceViewRenderer = new GLSurfaceViewRenderer(this);
    myGLSurfaceView.setRenderer(myGLSurfaceViewRenderer);
    // myGLSurfaceView.setRenderMode(GLSurfaceView.RENDERMODE_WHEN_DIRTY);
    setContentView(myGLSurfaceView);
  }

  @Override
  protected void onPause() {
    // Large objects in graphics memory should be de-allocated here
    // This is skipped here for simplicity of the example.
    super.onPause();
    myGLSurfaceView.onPause();
  }

  @Override
  protected void onResume() {
    super.onResume();
    // Large objects in graphics memory should be re-created here
    // if they were de-allocated on onPause().
    // This is skipped here for simplicity of the example.
    myGLSurfaceView.onResume();
  }
}
